@@ -8,6 +8,7 @@ This work can be distributed under the terms of the GNU GPLv3.
 
 from . import BUFSIZE
 from .database import NoSuchRowError
+from .backends.common import NoSuchObject
 from .multi_lock import MultiLock
 from .logging import logging # Ensure use of custom logger class
 from collections import OrderedDict
@@ -19,6 +20,7 @@ import hashlib
 import shutil
 import threading
 import time
+import sys
 
 # standard logger for this module
 log = logging.getLogger(__name__)
@@ -278,6 +280,9 @@ class BlockCache(object):
         if not os.path.exists(self.path):
             os.mkdir(self.path)
 
+        # Initialized fromt the outside to prevent cyclic dependency
+        self.fs = None
+
     def __len__(self):
         '''Get number of objects in cache'''
         return len(self.cache)
@@ -523,6 +528,7 @@ class BlockCache(object):
                 self._unlock_entry(el.inode, el.blockno)
                 raise
 
+        obj_lock_taken = False
         try:
             try:
                 old_block_id = self.db.get_val('SELECT block_id FROM inode_blocks '
@@ -551,6 +557,7 @@ class BlockCache(object):
 
                 with lock_released:
                     self._lock_obj(obj_id)
+                    obj_lock_taken = True
                     self._queue_upload((el, obj_id))
 
             # There is a block with the same hash
@@ -574,7 +581,8 @@ class BlockCache(object):
             self.in_transit.discard(el)
             with lock_released:
                 self._unlock_entry(el.inode, el.blockno, noerror=True)
-                self._unlock_obj(obj_id, noerror=True)
+                if obj_lock_taken:
+                    self._unlock_obj(obj_id)
             raise
 
         # Check if we have to remove an old block
@@ -682,10 +690,14 @@ class BlockCache(object):
             except QueueEmpty:
                 tmp = FlushSentinel
 
-            if (tmp is FlushSentinel or tmp is QuitSentinel) and ids:
+            if tmp in (FlushSentinel,QuitSentinel) and ids:
                 log.debug('removing: %s', ids)
-                with self.backend_pool() as backend:
-                    backend.delete_multi(['s3ql_data_%d' % i for i in ids])
+                try:
+                    with self.backend_pool() as backend:
+                        backend.delete_multi(['s3ql_data_%d' % i for i in ids])
+                except NoSuchObject:
+                    log.warning('Backend lost object s3ql_data_%d' % ids.pop(0))
+                    self.fs.failsafe = True
                 ids = []
             else:
                 ids.append(tmp)
@@ -693,7 +705,6 @@ class BlockCache(object):
             if tmp is QuitSentinel:
                 break
 
-
     @contextmanager
     def get(self, inode, blockno):
         """Get file handle for block `blockno` of `inode`
@@ -856,7 +867,7 @@ class BlockCache(object):
         This method releases the global lock.
         """
 
-        log.debug('started with %d, %d, %d', inode, start_no, end_no)
+        log.debug('started with %d, %d, %s', inode, start_no, end_no)
 
         if end_no is None:
             end_no = start_no + 1
@@ -930,19 +941,48 @@ class BlockCache(object):
         log.debug('finished')
 
     def get_usage(self):
-        '''Return cache size and dirty cache size
+        '''Get cache usage information.
+
+        Return a tuple of
+
+        * cache entries
+        * cache size
+        * dirty cache entries
+        * dirty cache size
+        * pending removals
 
         This method is O(n) in the number of cache entries.
         '''
 
         used = self.cache.size
-        dirty = 0
+        dirty_size = 0
+        dirty_cnt = 0
         for el in self.cache.values():
             if el.dirty:
-                dirty += el.size
+                dirty_size += el.size
+                dirty_cnt += 1
 
-        return (used, dirty)
+        if self.to_remove is None:
+            remove_cnt = 0
+        else:
+            # This is an estimate which may be negative
+            remove_cnt = max(0, self.to_remove.qsize())
+
+        return (len(self.cache), used, dirty_cnt, dirty_size, remove_cnt)
 
     def __del__(self):
-        if len(self.cache) > 0:
-            raise RuntimeError("BlockManager instance was destroyed without calling destroy()!")
+        # break reference loop
+        self.fs = None
+
+        if len(self.cache) == 0:
+            return
+
+        # Force execution of sys.excepthook (exceptions raised
+        # by __del__ are ignored)
+        try:
+            raise RuntimeError("BlockManager instance was destroyed without "
+                               "calling destroy()!")
+        except RuntimeError:
+            exc_info = sys.exc_info()
+
+        sys.excepthook(*exc_info)
