@@ -34,10 +34,10 @@ def parse_args(args):
     parser.add_debug()
     parser.add_quiet()
     parser.add_version()
-    parser.add_fatal_warnings()
 
     subparsers = parser.add_subparsers(metavar='<action>', dest='action',
                                        help='may be either of')
+    subparsers.required = True
     subparsers.add_parser('flushcache', help='flush file system cache',
                           parents=[pparser])
     subparsers.add_parser('upload-meta', help='Upload metadata',
@@ -89,11 +89,12 @@ def main(args=None):
         llfuse.setxattr(ctrlfile, 'upload-meta', b'dummy')
 
     elif options.action == 'log':
-        cmd = ('(%r, %r)' % (options.level, ','.join(options.modules))).encode()
+        level = getattr(logging, options.level.upper())
+        cmd = ('(%r, %r)' % (level, ','.join(options.modules))).encode()
         llfuse.setxattr(ctrlfile, 'logging', cmd)
 
     elif options.action == 'cachesize':
-        llfuse.setxattr(ctrlfile, 'cachesize', ('%d' % options.cachesize * 1024).encode())
+        llfuse.setxattr(ctrlfile, 'cachesize', ('%d' % (options.cachesize * 1024,)).encode())
 
 if __name__ == '__main__':
     main(sys.argv[1:])
