@@ -1,21 +1,22 @@
 '''
-mount.py - this file is part of S3QL (http://s3ql.googlecode.com)
+mount.py - this file is part of S3QL.
 
 Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>
 
-This program can be distributed under the terms of the GNU GPLv3.
+This work can be distributed under the terms of the GNU GPLv3.
 '''
 
 from .logging import logging, setup_logging, QuietError
-from . import fs, CURRENT_FS_REV, PICKLE_PROTOCOL
+from . import fs, CURRENT_FS_REV
 from .backends.pool import BackendPool
 from .block_cache import BlockCache
-from .common import (get_backend_cachedir, get_seq_no, stream_write_bz2, stream_read_bz2,
-                     get_backend_factory, pretty_print_size)
+from .common import (get_backend_cachedir, get_seq_no, get_backend_factory,
+                     load_params, save_params)
 from .daemonize import daemonize
 from .database import Connection
 from .inode_cache import InodeCache
-from .metadata import cycle_metadata, dump_metadata, restore_metadata
+from .metadata import (download_metadata, upload_metadata, dump_and_upload_metadata,
+                       dump_metadata)
 from .parse_args import ArgumentParser
 from .exit_stack import ExitStack
 from threading import Thread
@@ -27,7 +28,6 @@ import itertools
 import os
 import platform
 import subprocess
-import pickle
 import re
 import signal
 import resource
@@ -39,7 +39,7 @@ import shutil
 import atexit
 
 try:
-    import systemd.daemon.notify as sd_notify
+    from systemd.daemon import notify as sd_notify
 except ImportError:
     sd_notify = None
 
@@ -116,7 +116,8 @@ def main(args=None):
         import pstats
         prof = cProfile.Profile()
 
-    backend_factory = get_backend_factory(options)
+    backend_factory = get_backend_factory(options.storage_url, options.backend_options,
+                                          options.authfile, options.compress)
     backend_pool = BackendPool(backend_factory)
     atexit.register(backend_pool.flush)
 
@@ -138,7 +139,7 @@ def main(args=None):
         options.cachesize = min(rec_cachesize, 0.8 * avail_cache)
         log.info('Setting cache size to %d MB', options.cachesize / 1024)
     elif options.cachesize > avail_cache:
-        log.warning('Warning! Requested cache size %d MB, but only %d MB available',
+        log.warning('Requested cache size %d MB, but only %d MB available',
                     options.cachesize / 1024, avail_cache / 1024)
 
     if options.nfs:
@@ -157,10 +158,11 @@ def main(args=None):
     operations = fs.Operations(block_cache, db, max_obj_size=param['max_obj_size'],
                                inode_cache=InodeCache(db, param['inode_gen']),
                                upload_event=metadata_upload_thread.event)
+    block_cache.fs = operations
     metadata_upload_thread.fs = operations
 
     with ExitStack() as cm:
-        log.info('Mounting filesystem...')
+        log.info('Mounting %s at %s...', options.storage_url, options.mountpoint)
         try:
             llfuse.init(operations, options.mountpoint, get_fuse_opts(options))
         except RuntimeError as exc:
@@ -169,10 +171,7 @@ def main(args=None):
         unmount_clean = False
         def unmount():
             log.info("Unmounting file system...")
-            # Acquire lock so that Operations.destroy() is called with the
-            # global lock like all other handlers
-            with llfuse.lock:
-                llfuse.close(unmount=unmount_clean)
+            llfuse.close(unmount=unmount_clean)
         cm.callback(unmount)
 
         if options.fg:
@@ -207,10 +206,12 @@ def main(args=None):
             sd_notify('MAINPID=%d' % os.getpid())
 
         exc_info = setup_exchook()
+        workers = 1 if options.single else None # use default
+
         if options.profile:
-            prof.runcall(llfuse.main, options.single)
+            prof.runcall(llfuse.main, workers)
         else:
-            llfuse.main(options.single)
+            llfuse.main(workers)
 
         # Allow operations to terminate while block_cache is still available
         # (destroy() will be called again when from llfuse.close(), but at that
@@ -241,27 +242,11 @@ def main(args=None):
             log.info('File system unchanged, not uploading metadata.')
             del backend['s3ql_seq_no_%d' % param['seq_no']]
             param['seq_no'] -= 1
-            with open(cachepath + '.params', 'wb') as fh:
-                pickle.dump(param, fh, PICKLE_PROTOCOL)
+            save_params(cachepath, param)
         elif seq_no == param['seq_no']:
             param['last-modified'] = time.time()
-
-            log.info('Dumping metadata...')
-            with tempfile.TemporaryFile() as fh:
-                dump_metadata(db, fh)
-                def do_write(obj_fh):
-                    fh.seek(0)
-                    stream_write_bz2(fh, obj_fh)
-                    return obj_fh
-
-                log.info("Compressing and uploading metadata...")
-                obj_fh = backend.perform_write(do_write, "s3ql_metadata_new",
-                                               metadata=param, is_compressed=True)
-            log.info('Wrote %s of compressed metadata.', pretty_print_size(obj_fh.get_obj_size()))
-            log.info('Cycling metadata backups...')
-            cycle_metadata(backend)
-            with open(cachepath + '.params', 'wb') as fh:
-                pickle.dump(param, fh, PICKLE_PROTOCOL)
+            dump_and_upload_metadata(backend, db, param)
+            save_params(cachepath, param)
         else:
             log.error('Remote metadata is newer than local (%d vs %d), '
                       'refusing to overwrite!', seq_no, param['seq_no'])
@@ -358,8 +343,8 @@ def determine_threads(options):
         if threads > 0:
             log.info('Using %d upload threads (memory limited).', threads)
         else:
-            log.warning('Warning: compression will require %d MiB memory '
-                     '(%d%% of total system memory', mem_per_thread / 1024 ** 2,
+            log.warning('Compression will require %d MiB memory '
+                        '(%d%% of total system memory', mem_per_thread / 1024 ** 2,
                      mem_per_thread * 100 / memory)
             threads = 1
         return threads
@@ -376,8 +361,7 @@ def get_metadata(backend, cachepath):
     # Check for cached metadata
     db = None
     if os.path.exists(cachepath + '.params'):
-        with open(cachepath + '.params', 'rb') as fh:
-            param = pickle.load(fh)
+        param = load_params(cachepath)
         if param['seq_no'] < seq_no:
             log.info('Ignoring locally cached metadata (outdated).')
             param = backend.lookup('s3ql_metadata')
@@ -430,21 +414,8 @@ def get_metadata(backend, cachepath):
 
     # Download metadata
     if not db:
-        with tempfile.TemporaryFile() as tmpfh:
-            def do_read(fh):
-                tmpfh.seek(0)
-                tmpfh.truncate()
-                stream_read_bz2(fh, tmpfh)
-
-            log.info('Downloading and decompressing metadata...')
-            backend.perform_read(do_read, "s3ql_metadata")
-
-            log.info("Reading metadata...")
-            tmpfh.seek(0)
-            db = restore_metadata(tmpfh, cachepath + '.db')
-
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
+        db = download_metadata(backend, cachepath + '.db')
+    save_params(cachepath, param)
 
     return (param, db)
 
@@ -453,14 +424,7 @@ def mark_metadata_dirty(backend, cachepath, param):
 
     param['seq_no'] += 1
     param['needs_fsck'] = True
-    with open(cachepath + '.params', 'wb') as fh:
-        pickle.dump(param, fh, PICKLE_PROTOCOL)
-
-        # Fsync to make sure that the updated sequence number is committed to
-        # disk. Otherwise, a crash immediately after mount could result in both
-        # the local and remote metadata appearing to be out of date.
-        fh.flush()
-        os.fsync(fh.fileno())
+    save_params(cachepath, param)
     backend['s3ql_seq_no_%d' % param['seq_no']] = b'Empty'
     param['needs_fsck'] = False
 
@@ -468,7 +432,8 @@ def get_fuse_opts(options):
     '''Return fuse options for given command line options'''
 
     fuse_opts = [ "nonempty", 'fsname=%s' % options.storage_url,
-                  'subtype=s3ql' ]
+                  'subtype=s3ql', 'big_writes', 'max_write=131072',
+                  'no_remote_lock' ]
 
     if platform.system() == 'Darwin':
         # FUSE4X and OSXFUSE claim to support nonempty, but
@@ -541,7 +506,6 @@ def parse_args(args):
     parser.add_backend_options()
     parser.add_version()
     parser.add_storage_url()
-    parser.add_fatal_warnings()
 
     parser.add_argument("mountpoint", metavar='<mountpoint>', type=os.path.abspath,
                         help='Where to mount the file system')
@@ -666,7 +630,7 @@ class MetadataUploadThread(Thread):
 
             with self.backend_pool() as backend:
                 seq_no = get_seq_no(backend)
-                if seq_no != self.param['seq_no']:
+                if seq_no > self.param['seq_no']:
                     log.error('Remote metadata is newer than local (%d vs %d), '
                               'refusing to overwrite and switching to failsafe mode!',
                               seq_no, self.param['seq_no'])
@@ -679,17 +643,7 @@ class MetadataUploadThread(Thread):
 
                 # Temporarily decrease sequence no, this is not the final upload
                 self.param['seq_no'] -= 1
-                def do_write(obj_fh):
-                    fh.seek(0)
-                    stream_write_bz2(fh, obj_fh)
-                    return obj_fh
-                log.info("Compressing and uploading metadata...")
-                obj_fh = backend.perform_write(do_write, "s3ql_metadata_new",
-                                               metadata=self.param, is_compressed=True)
-                log.info('Wrote %s of compressed metadata.',
-                         pretty_print_size(obj_fh.get_obj_size()))
-                log.info('Cycling metadata backups...')
-                cycle_metadata(backend)
+                upload_metadata(backend, fh, self.param)
                 self.param['seq_no'] += 1
 
                 fh.close()
