@@ -20,8 +20,9 @@ from s3ql.block_cache import BlockCache, QuitSentinel
 from s3ql.mkfs import init_tables
 from s3ql.metadata import create_tables
 from s3ql.database import Connection
-from s3ql.common import AsyncFn
-from common import catch_logmsg, safe_sleep
+from s3ql.common import AsyncFn, time_ns
+from common import safe_sleep
+from pytest_checklogs import assert_logs
 import llfuse
 import errno
 import os
@@ -30,7 +31,6 @@ import shutil
 import stat
 import tempfile
 import threading
-import time
 import unittest
 import queue
 import pytest
@@ -61,6 +61,9 @@ class DummyQueue:
             self.obj = QuitSentinel
             return tmp
 
+    def qsize(self):
+        return 0
+
 class cache_tests(unittest.TestCase):
 
     def setUp(self):
@@ -82,11 +85,12 @@ class cache_tests(unittest.TestCase):
 
         # Create an inode we can work with
         self.inode = 42
-        self.db.execute("INSERT INTO inodes (id,mode,uid,gid,mtime,atime,ctime,refcount,size) "
+        now_ns = time_ns()
+        self.db.execute("INSERT INTO inodes (id,mode,uid,gid,mtime_ns,atime_ns,ctime_ns,refcount,size) "
                         "VALUES (?,?,?,?,?,?,?,?,?)",
                         (self.inode, stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
                          | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH,
-                         os.getuid(), os.getgid(), time.time(), time.time(), time.time(), 1, 32))
+                         os.getuid(), os.getgid(), now_ns, now_ns, now_ns, 1, 32))
 
         cache = BlockCache(self.backend_pool, self.db, self.cachedir + "/cache",
                            self.max_obj_size * 100)
@@ -160,7 +164,7 @@ class cache_tests(unittest.TestCase):
         try:
             # Try to clean-up (implicitly calls expire)
             with llfuse.lock_released, \
-                catch_logmsg('Unable to flush cache, no upload threads left alive',
+                assert_logs('Unable to flush cache, no upload threads left alive',
                               level=logging.ERROR, count=1):
                 with pytest.raises(OSError) as exc_info:
                      self.cache.destroy()
@@ -227,7 +231,7 @@ class cache_tests(unittest.TestCase):
 
         # We want to expire 4 entries, 2 of which are already flushed
         self.cache.cache.max_entries = 16
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=2)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=2)
         self.cache.expire()
         self.cache.backend_pool.verify()
         self.assertEqual(len(self.cache.cache), 16)
@@ -250,7 +254,7 @@ class cache_tests(unittest.TestCase):
         data3 = self.random_data(datalen)
 
         # Case 1: create new object
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         with self.cache.get(inode, blockno1) as fh:
             fh.seek(0)
             fh.write(data1)
@@ -259,7 +263,7 @@ class cache_tests(unittest.TestCase):
         self.cache.backend_pool.verify()
 
         # Case 2: Link new object
-        self.cache.backend_pool = TestBackendPool(self.backend_pool)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool)
         with self.cache.get(inode, blockno2) as fh:
             fh.seek(0)
             fh.write(data1)
@@ -268,7 +272,7 @@ class cache_tests(unittest.TestCase):
         self.cache.backend_pool.verify()
 
         # Case 3: Upload old object, still has references
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         with self.cache.get(inode, blockno1) as fh:
             fh.seek(0)
             fh.write(data2)
@@ -276,7 +280,7 @@ class cache_tests(unittest.TestCase):
         self.cache.backend_pool.verify()
 
         # Case 4: Upload old object, no references left
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_del=1, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_del=1, no_write=1)
         with self.cache.get(inode, blockno2) as fh:
             fh.seek(0)
             fh.write(data3)
@@ -284,7 +288,7 @@ class cache_tests(unittest.TestCase):
         self.cache.backend_pool.verify()
 
         # Case 5: Link old object, no references left
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_del=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_del=1)
         with self.cache.get(inode, blockno2) as fh:
             fh.seek(0)
             fh.write(data2)
@@ -293,7 +297,7 @@ class cache_tests(unittest.TestCase):
 
         # Case 6: Link old object, still has references
         # (Need to create another object first)
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         with self.cache.get(inode, blockno3) as fh:
             fh.seek(0)
             fh.write(data1)
@@ -301,7 +305,7 @@ class cache_tests(unittest.TestCase):
         self.cache.upload(el3)
         self.cache.backend_pool.verify()
 
-        self.cache.backend_pool = TestBackendPool(self.backend_pool)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool)
         with self.cache.get(inode, blockno1) as fh:
             fh.seek(0)
             fh.write(data1)
@@ -316,7 +320,7 @@ class cache_tests(unittest.TestCase):
         blockno2 = 24
         data = self.random_data(datalen)
 
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         with self.cache.get(inode, blockno1) as fh:
             fh.seek(0)
             fh.write(data)
@@ -326,7 +330,7 @@ class cache_tests(unittest.TestCase):
         self.cache.clear()
         self.cache.backend_pool.verify()
 
-        self.cache.backend_pool = TestBackendPool(self.backend_pool)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool)
         self.cache.remove(inode, blockno1)
         self.cache.backend_pool.verify()
 
@@ -446,10 +450,10 @@ class cache_tests(unittest.TestCase):
         with self.cache.get(inode, 1) as fh:
             fh.seek(0)
             fh.write(data1)
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         commit(self.cache, inode)
         self.cache.backend_pool.verify()
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_del=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_del=1)
         self.cache.remove(inode, 1)
         self.cache.backend_pool.verify()
 
@@ -466,10 +470,10 @@ class cache_tests(unittest.TestCase):
         with self.cache.get(inode, 1) as fh:
             fh.seek(0)
             fh.write(data1)
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_write=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_write=1)
         self.cache.clear()
         self.cache.backend_pool.verify()
-        self.cache.backend_pool = TestBackendPool(self.backend_pool, no_del=1)
+        self.cache.backend_pool = MockBackendPool(self.backend_pool, no_del=1)
         self.cache.remove(inode, 1)
         self.cache.backend_pool.verify()
         with self.cache.get(inode, 1) as fh:
@@ -477,7 +481,7 @@ class cache_tests(unittest.TestCase):
             self.assertEqual(fh.read(42), b'')
 
 
-class TestBackendPool(AbstractBackend):
+class MockBackendPool(AbstractBackend):
     has_native_rename = False
 
     def __init__(self, backend_pool, no_read=0, no_write=0, no_del=0):
